# absolute imports
from PySide6.QtWidgets import (
    QVBoxLayout,
    QLabel,
    QWidget,
    QGroupBox,
    QPushButton,
    QCheckBox,
    QLineEdit,
    QHBoxLayout,
    QColorDialog,
    QComboBox,
    QTabWidget,
    QPlainTextEdit,
    QStyleFactory,
    QDialog,
)
from PySide6.QtGui import QIcon, QGuiApplication, QColor, QTextCursor, QImage, QPixmap,\
    qRed, qGreen, qBlue, qAlpha, qRgba
from PySide6.QtCore import Qt, Signal
import numpy as np


# relative imports
from . import __version__ as measury_version
from . import measury_path
from .settings import Settings, DEFAULT_SETTINGS


class MeasuryWindow(QWidget):
    """
    This "window" is a QWidget. If it has no parent, it
    will appear as a free-floating window as we want.
    """

    def __init__(self, parent, title="Window"):
        super().__init__()
        # this makes the main application unresponsive while using this window
        # self.setWindowModality(Qt.WindowModality.ApplicationModal)

        self.parent = parent

        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

        self.setWindowTitle(f"Measury: {title}")
        self.setWindowIcon(QIcon(str(measury_path / "data/tape_measure_128.ico")))
        self.setMinimumSize(300, 200)

    def closeEvent(self, event):
        """
        Override the close event to set focus back to the parent window.
        """
        if self.parent:
            self.parent.setFocus()
            self.parent.activateWindow()
        super().closeEvent(event)


class SaveWindow(MeasuryWindow):
    """
    The window displayed to save the data.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(title="Save Dialog", *args, **kwargs)

        # Save Options
        self.options_box = QGroupBox("Options", self)
        self.options_layout = QVBoxLayout()
        self.options_box.setLayout(self.options_layout)
        self.layout.addWidget(self.options_box)

        self.save_img_checkbox = QCheckBox(text="Save Image")
        self.save_img_checkbox.setChecked(True)
        self.options_layout.addWidget(self.save_img_checkbox)

        # Save Button
        self.saveButton = QPushButton("Save", self)
        self.saveButton.clicked.connect(self.save)
        self.layout.addWidget(self.saveButton)

    def save(self):
        return self.parent.data_handler.save_storage_file(
            save_image=self.save_img_checkbox.isChecked()
        )


class AboutWindow(MeasuryWindow):
    """
    The window displaying the about.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(title="About", *args, **kwargs)

        self.info_layout = QVBoxLayout()

        self.info_layout.addWidget(QLabel(f"<b>Measury {measury_version}</b>"))
        self.info_layout.addWidget(
            QLabel("A simple tool for measuring distances in images.")
        )
        self.info_layout.addWidget(QLabel("Developed by:"))
        self.info_layout.addWidget(QLabel("Jan Ullmann"))

        self.layout.addLayout(self.info_layout)


class DataWindow(MeasuryWindow):
    """
    The window displaying the data results.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(title="Measurements", *args, **kwargs)

        label = QLabel(self.parent.data_handler.calculate_results_string())
        label.setTextInteractionFlags(
            Qt.TextInteractionFlag.TextSelectableByMouse
        )  # Make the text selectable
        self.layout.addWidget(label)

        # Copy Button
        self.copyButton = QPushButton("Copy to Clipboard", self)
        self.copyButton.clicked.connect(self.copy_to_clipboard)
        self.layout.addWidget(self.copyButton)
        # Save Button
        self.saveButton = QPushButton("Save", self)
        self.saveButton.clicked.connect(self.parent.data_handler.save_measurements_file)
        self.layout.addWidget(self.saveButton)

    def copy_to_clipboard(self):
        cb = QGuiApplication.clipboard()
        cb.clear()
        cb.setText(self.parent.data_handler.calculate_results_string())

class ImageWindow(QDialog):
    """
    The window displaying the image.
    """
    image = None

    def __init__(self, image:QImage, title="Image Editor", *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        self.image = image
        self.orig_image = self.extract_img_arr(image)[1]

        self.setWindowTitle(title)
        self.setWindowIcon(QIcon(str(measury_path / "data/tape_measure_128.ico")))
        self.setMinimumSize(300, 200)
        self.setMaximumSize(800, 600)

        # force this window on top
        self.setWindowModality(Qt.WindowModality.ApplicationModal)

        # Create a QVBoxLayout
        layout = QVBoxLayout()

        # Create a QLabel and set the QImage as its pixmap
        self.label = QLabel()
        self.label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.redraw_image()

        # Add the QLabel to the layout
        layout.addWidget(self.label)

        # add buttons
        self.button_layout = QHBoxLayout()

        self.conversion_dd = QComboBox(self)
        self.conversion_dd.addItems(["RGB", "RBG", "GRB", "GBR", "BRG", "BGR"])
        self.conversion_dd.currentTextChanged.connect(self.swap_colors)
        layout.addWidget(self.conversion_dd)

        self.reset_button = QPushButton("Reset", self)
        self.button_layout.addWidget(self.reset_button)
        self.reset_button.clicked.connect(self.reset_image)

        self.save_button = QPushButton("Save", self)
        self.button_layout.addWidget(self.save_button)
        self.save_button.clicked.connect(self.close)

        # set focus on save so that enter directly closes the window
        self.save_button.setFocus()

        # Set the layout to the window
        layout.addLayout(self.button_layout)
        self.setLayout(layout)

    def extract_img_arr(self, image:QImage) -> np.ndarray:
        image = image.convertToFormat(QImage.Format.Format_RGBA8888)
        width = image.width()
        height = image.height()

        # Get the pixel data as a NumPy array
        ptr = image.bits()
        return image, np.array(ptr).reshape((height, width, 4))  # 4 channels: R, G, B, A

    def swap_colors(self):
        # Convert QImage to a format that allows direct access to pixel data
        color_order = self.conversion_dd.currentText()
        image = self.image.convertToFormat(QImage.Format.Format_RGBA8888)
        arr = self.orig_image.copy() 

        # Define the channel order
        channel_map = {
            "RGB": [0, 1, 2],
            "RBG": [0, 2, 1],
            "GRB": [1, 0, 2],
            "GBR": [1, 2, 0],
            "BRG": [2, 0, 1],
            "BGR": [2, 1, 0],
        }

        # Swap the color channels
        arr[..., :3] = arr[..., channel_map[color_order]]

        # Convert the NumPy array back to QImage
        self.image = QImage(arr.data, arr.shape[1], arr.shape[0], image.bytesPerLine(), QImage.Format.Format_RGBA8888)
        self.redraw_image()

    def reset_image(self):
        self.conversion_dd.setCurrentIndex(0)
        self.redraw_image() 
        
    def redraw_image(self):
        scaled_pixmap = QPixmap.fromImage(self.image).scaled(
            self.label.size(), Qt.AspectRatioMode.KeepAspectRatio, Qt.TransformationMode.SmoothTransformation
        )
        self.label.setPixmap(scaled_pixmap)

    def resizeEvent(self, event):
        # Call the redraw_image method to scale the image when the window is resized
        self.redraw_image()
        super().resizeEvent(event)


class SettingsWindow(MeasuryWindow):

    def __init__(self, *args, **kwargs):
        super().__init__(title="Settings", *args, **kwargs)

        self.settings: Settings = self.parent.settings
        self.settings.load_settings()

        self.main_layout = QVBoxLayout()
        self.tab_widget = QTabWidget()
        self.main_layout.addWidget(self.tab_widget)

        self.graphics = QWidget()
        self.ui = QWidget()
        self.misc = QWidget()

        self.tab_widget.addTab(self.graphics, "Graphics")
        self.tab_widget.addTab(self.ui, "User Interface")
        self.tab_widget.addTab(self.misc, "Misc")
        self.tab_widget.setMinimumSize(350, 200)

        self.create_graphics_ui()
        self.create_ui_ui()
        self.create_misc_ui()

        # Resize the window to fit the contents of the tabs
        self.resize(self.tab_widget.sizeHint())

        # Buttons Layout
        self.button_layout = QHBoxLayout()
        self.main_layout.addLayout(self.button_layout)

        # Save Button
        self.save_button = QPushButton("Save", self)
        self.save_button.clicked.connect(self.save)
        self.save_button.setEnabled(self.changed)
        self.button_layout.addWidget(self.save_button)

        # reset button
        self.reset_button = QPushButton("Reset to defaults", self)
        self.reset_button.clicked.connect(self.reset_to_defaults)
        self.reset_button.setEnabled(not self.settings.is_default)
        self.button_layout.addWidget(self.reset_button)

        self.layout.addLayout(self.main_layout)
        self.adjustSize()
        
    def create_ui_ui(self):
        
        self.ui_layout = QVBoxLayout()
        self.ui.setLayout(self.ui_layout)

        self.default_microscope_layout = QHBoxLayout()
        self.default_microscope_label = QLabel("Default Microscope", self)
        self.default_microscope_layout.addWidget(self.default_microscope_label)
        self.default_microscope_dd = QComboBox(self)
        self.default_microscope_dd.addItems(self.parent.data_handler.micros_db.keys())
        self.default_microscope_dd.setCurrentText(self.settings.value("ui/microscope"))
        self.default_microscope_dd.currentTextChanged.connect(self.update_window)
        self.default_microscope_layout.addWidget(self.default_microscope_dd)
        self.ui_layout.addLayout(self.default_microscope_layout)
        
        self.show_both_scaling_layout = QHBoxLayout()
        self.show_both_scaling_label = QLabel("Show Scaling in px and unit", self)
        self.show_both_scaling_layout.addWidget(self.show_both_scaling_label)
        self.show_both_scaling_cb = QCheckBox(self)
        self.show_both_scaling_cb.setChecked(self.settings.value("ui/show_both_scaling", type=bool))
        self.show_both_scaling_cb.stateChanged.connect(self.update_window)
        self.show_both_scaling_layout.addWidget(self.show_both_scaling_cb)
        self.ui_layout.addLayout(self.show_both_scaling_layout)

        self.ui_layout.addStretch()

    def create_misc_ui(self):

        self.misc_layout = QVBoxLayout()
        self.misc.setLayout(self.misc_layout)

        self.clearUndoHistButton = QPushButton("Clear Undo History", self)
        self.clearUndoHistButton.clicked.connect(self.parent.reset_undo_stack)
        self.misc_layout.addWidget(self.clearUndoHistButton)

        self.file_extensions_layout = QHBoxLayout()
        self.file_extensions_label = QLabel("File Extensions", self)
        self.file_extensions_layout.addWidget(self.file_extensions_label)
        self.file_extensions = QLineEdit(self)
        self.file_extensions.setPlaceholderText(".msry;.measury")
        file_ext_string = ";".join(self.settings.value("misc/file_extensions"))
        self.file_extensions.setText(file_ext_string)
        self.file_extensions.textChanged.connect(self.update_window)
        self.file_extensions_layout.addWidget(self.file_extensions)
        self.misc_layout.addLayout(self.file_extensions_layout)

        self.misc_layout.addStretch()

    def create_graphics_ui(self):
        """Create the UI for the settings window."""

        # rendering version
        self.graphics_layout = QVBoxLayout()

        self.rendering_layout = QHBoxLayout()
        self.rendering_label = QLabel("Image Rendering", self)

        self.rendering_combobox = QComboBox(self)
        self.rendering_combobox.addItems(
            [
                "bessel",
                "blackman",
                "catrom",
                "cubic",
                "custom",
                "gaussian",
                "hamming",
                "hanning",
                "hermite",
                "kaiser",
                "lanczos",
                "linear",
                "mitchell",
                "nearest",
                "quadric",
                "sinc",
                "spline16",
                "spline36",
                "bilinear",
                "bicubic",
            ]
        )
        self.rendering_combobox.setCurrentText(
            self.settings.value("graphics/image_rendering")
        )
        self.rendering_combobox.currentTextChanged.connect(self.update_window)

        self.rendering_layout.addWidget(self.rendering_label)
        self.rendering_layout.addWidget(self.rendering_combobox)
        self.graphics_layout.addLayout(self.rendering_layout)

        # graphics style
        self.graphics_style_layout = QHBoxLayout()
        self.graphics_style_label = QLabel("Graphics Style", self)

        self.graphics_style_dd = QComboBox(self)

        available_styles = QStyleFactory.keys()

        self.graphics_style_dd.addItems(available_styles)
        self.graphics_style_dd.setCurrentText(self.settings.value("graphics/style"))
        self.graphics_style_dd.currentTextChanged.connect(self.update_window)

        self.graphics_style_layout.addWidget(self.graphics_style_label)
        self.graphics_style_layout.addWidget(self.graphics_style_dd)
        self.graphics_layout.addLayout(self.graphics_style_layout)

        # Color Selector inside
        self.color_layout = QHBoxLayout()
        self.color_label = QLabel("Object Color", self)

        self.color_picker = ColorPicker(
            self.settings.value("graphics/object_color"), "Pick Color", self
        )
        self.color_picker.color_updated.connect(self.update_window)

        self.color_layout.addWidget(self.color_label)
        self.color_layout.addWidget(self.color_picker)
        self.graphics_layout.addLayout(self.color_layout)

        # Color Selector border
        self.border_color_layout = QHBoxLayout()
        self.border_color_label = QLabel("Object Border Color", self)

        self.border_color_picker = ColorPicker(
            self.settings.value("graphics/object_border_color"), "Pick Color", self
        )
        self.border_color_picker.color_updated.connect(self.update_window)

        self.border_color_layout.addWidget(self.border_color_label)
        self.border_color_layout.addWidget(self.border_color_picker)
        self.graphics_layout.addLayout(self.border_color_layout)

        # Color Selector scale bar
        self.scale_bar_color_layout = QHBoxLayout()
        self.scale_bar_color_label = QLabel("Scale Bar Color", self)

        self.scale_bar_color_picker = ColorPicker(
            self.settings.value("graphics/scale_bar_color"),
            "Pick Color",
            self,
            alpha_channel=False,
        )
        self.scale_bar_color_picker.color_updated.connect(self.update_window)

        self.scale_bar_color_layout.addWidget(self.scale_bar_color_label)
        self.scale_bar_color_layout.addWidget(self.scale_bar_color_picker)
        self.graphics_layout.addLayout(self.scale_bar_color_layout)

        self.graphics_layout.addStretch()

        self.graphics.setLayout(self.graphics_layout)

    def set_settings(self, settings: dict):
        """Set Settings to a value."""

        self.rendering_combobox.setCurrentText(settings.get("graphics/image_rendering"))
        self.color_picker.selectedColor = settings.get("graphics/object_color")
        self.border_color_picker.selectedColor = settings.get(
            "graphics/object_border_color"
        )
        self.scale_bar_color_picker.selectedColor = settings.get(
            "graphics/scale_bar_color"
        )
        self.graphics_style_dd.setCurrentText(settings.get("graphics/style"))

        self.default_microscope_dd.setCurrentText(settings.get("ui/microscope"))
        self.show_both_scaling_cb.setChecked(settings.get("ui/show_both_scaling"))
        
        file_ext_string = ";".join(settings.get("misc/file_extensions"))
        self.file_extensions.setText(file_ext_string)
        self.update_window()

    def reset_to_defaults(self):
        """Reset the settings to the default values."""
        self.set_settings(DEFAULT_SETTINGS)
        self.reset_button.setEnabled(False)

    def update_window(self):
        """Update the windows with the new settings."""
        self.parent.data_handler.logger.info("Updating settings window")
        # set button states
        self.reset_button.setEnabled(not self.settings.is_default)
        self.save_button.setEnabled(self.changed)

    @property
    def current_selection(self):
        return {
            "graphics/object_color": self.color_picker.selectedColor,
            "graphics/object_border_color": self.border_color_picker.selectedColor,
            "graphics/image_rendering": self.rendering_combobox.currentText(),
            "graphics/scale_bar_color": self.scale_bar_color_picker.selectedColor,
            "graphics/style": self.graphics_style_dd.currentText(),
            "ui/microscope": self.default_microscope_dd.currentText(),
            "ui/show_both_scaling": self.show_both_scaling_cb.isChecked(),
            "misc/file_extensions": self.file_extensions.text().split(";"),
        }

    def save(self):
        for key in self.current_selection.keys():
            if key in DEFAULT_SETTINGS.keys():
                value = self.current_selection.get(key)
                self.settings.save(key, value)

        # update the window and the color palette
        self.parent.update_style()

        self.update_window()
        # update color of the scalebar if it has been drawn
        if self.parent.vispy_canvas.scale_bar_params[0] is not None:
            self.parent.vispy_canvas.find_scale_bar_width(
                *self.parent.vispy_canvas.scale_bar_params
            )
        else:
            self.parent.vispy_canvas.draw_image()
        self.parent.vispy_canvas.update_colors()
        self.parent.right_ui.update_colors()
        self.parent.main_ui.set_selected_object_table_columns()
        self.parent.main_ui.update_object_table()

    @property
    def changed(self):
        """Check if there is a change in the settings gui."""
        return not self.settings.equals_settings(self.current_selection)


class ColorPicker(QPushButton):

    color_updated = Signal(QColor)  # Define a custom signal

    def __init__(self, color: QColor, *args, alpha_channel=True, **kwargs):
        super().__init__(*args, **kwargs)
        self.color_picker = QColorDialog()
        if alpha_channel:
            self.color_picker.setOption(QColorDialog.ColorDialogOption.ShowAlphaChannel)
        self.color_picker.colorSelected.connect(self.update_color)
        self.clicked.connect(self.color_picker.open)
        self.selectedColor = color  # Initialize with a default color

    def update_color(self, color):
        self._selectedColor = color

        rgba = color.getRgb()
        self.setStyleSheet(f"background-color: rgba{rgba};")
        self.color_updated.emit(color)

    @property
    def selectedColor(self) -> QColor:
        return self._selectedColor

    @selectedColor.setter
    def selectedColor(self, color: QColor):
        self.color_picker.setCurrentColor(color)
        self.update_color(color)


class XMLWindow(MeasuryWindow):
    """
    The window displaying the xml values of a file.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(title="XML file data", *args, **kwargs)

        self.setMinimumHeight(500)

        byte_stream = self.parent.data_handler.img_byte_stream

        values = self.parent.main_ui.get_microscope().get_metadata(byte_stream)
        # create scrollable text area
        self.text_area = QPlainTextEdit(self)
        self.text_area.setPlainText(values)
        self.text_area.setReadOnly(True)

        # create search line edit
        self.search_line_edit = QLineEdit(self)
        self.search_line_edit.setPlaceholderText("Search")
        self.search_line_edit.returnPressed.connect(self.search_text)

        # create search button
        self.search_button = QPushButton("Search", self)
        self.search_button.clicked.connect(self.search_text)

        # add widgets to layout
        self.search_layout = QHBoxLayout()
        self.search_layout.addWidget(self.search_line_edit)
        self.search_layout.addWidget(self.search_button)

        self.layout.addLayout(self.search_layout)
        self.layout.addWidget(self.text_area)

    def search_text(self):
        # get search query from line edit
        query = self.search_line_edit.text()

        # search text area for query
        found = self.text_area.find(query)

        if not found:
            # if query was not found, move cursor to start and search again
            self.text_area.moveCursor(QTextCursor.MoveOperation.Start)
            self.text_area.find(query)



import PyInstaller.__main__
import shutil
import sys
import platform
from pathlib import Path
from measury import __version__ as measury_version

script_path = Path(__file__).parent.resolve()

def compile(program_name, method="onedir", zipped=False):
    # run pyinstaller
    arguments = [
            str(script_path / "run_clean.py"),
            f"--{method}",
            "--windowed",
            "--name",
            program_name,
            "--specpath",
            "build",
            "--collect-all",
            "vispy",
            "--collect-all",
            "measury",
            "--collect-all",
            "scipy",
            "--noconfirm",
            "--icon",
        ]
    
    if sys.platform == "darwin":
        arguments.append("../img/logo/tape_measure_128.icns") 
        arguments.append("--windowed")
        extension = ".app"

    else:
        arguments.append("../img/logo/tape_measure_128.ico")
        extension = ""


    PyInstaller.__main__.run(
        arguments
    )

    # zip output for easy distribution
    output_path = (script_path / f"../dist/{program_name}{extension}").resolve()
    # Modify the output_path based on architecture
    if zipped:

        # Determine the architecture
        arch = platform.machine()
        match sys.platform:
            case "win32":
                archive_format = "zip"
                suffix = "_win"
            case "linux":
                archive_format = "gztar"
                suffix = "_lnx"
            case "darwin":
                archive_format = "gztar"
                suffix = "_mac"

        # Append architecture to suffix
        if "arm" in arch.lower():
            suffix += "_arm"
        elif "x86_64" in arch.lower() or "amd64" in arch.lower():
            suffix += "_x64"

        shutil.make_archive(
            output_path.with_name(f"{program_name}{suffix}{extension}"),
            archive_format,
            root_dir=output_path.parent, # the dist folder so that we can archive the full folder
            base_dir=f"{program_name}{extension}",
        )
        
if __name__ == "__main__":
    # set program name and method
    program_name = "Measury" + "_" + measury_version
    method = "onedir"
    zipped = False
    if len(sys.argv) > 1 and sys.argv[1] == "zipped":
        zipped = True
        
    compile(program_name, method, zipped)



import measury
import logging
from pathlib import Path

script_dir = Path(__file__).parent.resolve()

if __name__ == "__main__":
    logging.basicConfig(
        level=logging.CRITICAL,
        format="%(asctime)s  %(levelname)-10s %(name)s: %(message)s",
    )
    logger = logging.getLogger("Measury")
    logger.setLevel(logging.DEBUG)
    measury.run(logger=logger)
    # measury.run(file_path=script_dir/r"tests/test_data/test_image.tif", logger=logger)
    # measury.run(file_path=script_dir/r"tests/test_data/test_file_4.msry", logger=logger)


import inspect
import sys
from abc import ABC
from typing import Optional, Tuple
import json
import xml.etree.ElementTree as ET
import io


def load_microscopes():
    microscopes = dict()
    for name, obj in inspect.getmembers(sys.modules[__name__]):
        if inspect.isclass(obj) and issubclass(obj, Microscope) and obj != Microscope:
            microscopes[name] = obj
    return microscopes


class Microscope(ABC):
    def __init__(
        self,
        seed_points: Optional[Tuple[float, float]] = None,
        orientation: Optional[str] = None,
        threshold: Optional[int] = None,
    ):
        """
        Initializes a new instance of the Microscope class.

        :param seed_points: A tuple representing the seed points for the
                            microscope.
        :param orientation: The orientation of the microscope.
        :param threshold: The threshold value used for processing.
        """
        self.seed_points = seed_points
        self.orientation = orientation
        self.threshold = threshold

    def get_metadata(self, byte_stream: Optional[bytes] = None) -> str:
        """
        Abstract method to get the metadata of a microscope image.

        :param file_path: The path to the file from which to extract metadata.
        :return: A string representing the metadata.
        """
        if byte_stream is None:
            return ""
        else:
            return byte_stream.decode("utf-8", errors="ignore")


class Generic_Microscope(Microscope):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Zeiss_Orion_Nanofab(Microscope):

    def __init__(self):
        super().__init__(
            seed_points=(0.698243, 0.92768),
            orientation="horizontal",
            threshold=10
        )

    def get_metadata(self, byte_stream=None) -> str:

        values = self.get_xml(byte_stream)

        # check if values is a non-empty dictionary
        if isinstance(values, dict):
            return json.dumps(values, indent=4, ensure_ascii=False)
        return Microscope().get_metadata(byte_stream)

    def get_xml(self, byte_stream=None) -> dict:
        # This only works for Zeiss Orion files yet
        # first check if image was loaded
        if byte_stream is None:
            return None

        # get last line of file
        last_line = get_last_line(byte_stream)

        try:
            root = ET.fromstring(last_line)
        except ET.ParseError as e:
            # return None
            raise Exception(
                "There was a problem parsing the XML string.\n"
                "Problematic part of the XML string: "
                f"'{last_line[max(0, e.position[1]-10):e.position[1]+10]}' "
                "Are you sure this is a Zeiss Orion Nanofab file?"
            )
        # Create a dictionary to store the values
        values = parse_element(root)

        return values


def get_last_line(byte_stream) -> str:
    with io.BytesIO(byte_stream) as f:
        # Go to the end of the file
        f.seek(0, io.SEEK_END)
        # Get the position (this is the size of the file in bytes)
        position = f.tell()
        line = b""
        # we're now moving backwards through the file
        while position >= 0:
            f.seek(position)
            next_char = f.read(1)
            # Stop if a newline is found and line is not empty
            if next_char == b"\n" and line:
                break
            line = next_char + line
            position -= 1
        # remove last end character
        last_line = line.decode(encoding="utf-8", errors="replace").strip()[:-1]
        # remove all elements before the first '<' and the last '>' character
        last_line = last_line[last_line.find("<"):last_line.rfind(">") + 1]
        last_line.replace("\n", "")
        
        return last_line


def parse_element(element) -> dict:
    """Recursively parse an XML element and its children into a dictionary."""

    # Create a dictionary to store the element's attributes
    node = {}

    # Iterate over all child elements
    for child in list(element):
        # Recursively parse the child element
        child_data = parse_element(child)

        # If the child's tag is already in the node dictionary,
        # add the child's data to that tag
        if child.tag in node:
            # If there's only one child with this tag so far, put it in a list
            if type(node[child.tag]) is list:
                node[child.tag].append(child_data)
            else:
                node[child.tag] = [node[child.tag], child_data]
        else:
            node[child.tag] = child_data

    # If the element has no child elements, store its text
    if not node:
        node = element.text

    return node


class Zeiss_Crossbeam_550L(Microscope):

    def __init__(self):
        super().__init__(
            seed_points=(0.02051, 0.94922),
            orientation="horizontal",
            threshold=10
        )


class Jeol_JSM_6500F(Microscope):

    def __init__(self):
        super().__init__(
            seed_points=(0.777344, 0.951172),
            orientation="horizontal",
            threshold=10
        )


if __name__ == "__main__":
    microscopes = load_microscopes()
    for microscope in microscopes:
        print(microscope)



# absolute imports
from PySide6.QtWidgets import (
    QVBoxLayout,
    QHBoxLayout,
    QPushButton,
    QTableWidget,
    QWidget,
    QFileDialog,
    QComboBox,
    QLabel,
    QButtonGroup,
    QLineEdit,
    QGroupBox,
    QTableWidgetItem,
    QListWidget,
    QInputDialog,
    QHeaderView,
)
from PySide6.QtGui import QDoubleValidator, QIntValidator, QFontMetrics, QUndoCommand
from PySide6.QtCore import Qt
from pathlib import Path
import numpy as np

# relative imports
from .windows import SaveWindow
from .data.microscopes import Microscope


class MainUI(QWidget):

    main_window = None
    update_full_table = False
    DEFAULT_THRESHOLD = 10

    def __init__(self, main_window, parent=None):

        self.parent = parent
        super().__init__(self.parent)

        self.main_window = main_window
        self.vispy_canvas = main_window.vispy_canvas
        self.data_handler = main_window.data_handler
        self.right_ui = main_window.right_ui

        self.layout = QVBoxLayout(self)
        self.setMinimumWidth(250)

        # image settings
        self.image_box = QGroupBox("Image Settings", self)
        self.image_layout = QHBoxLayout()

        # self.select_image_button = QPushButton("Select File", self)
        # self.image_layout.addWidget(self.select_image_button)
        # self.select_image_button.clicked.connect(
        #     lambda: self.select_file(file_path=None)
        # )

        self.center_image_button = QPushButton("Center Image", self)
        self.image_layout.addWidget(self.center_image_button)
        self.center_image_button.clicked.connect(self.vispy_canvas.center_image)

        image_tools = dict()
        image_tools["set_origin"] = QPushButton("Set Origin", self)
        self.image_layout.addWidget(image_tools["set_origin"])

        self.image_box.setLayout(self.image_layout)
        self.layout.addWidget(self.image_box)

        # tools settings
        self.tools_edit_box = QGroupBox("Edit", self)
        self.tools_edit_layout = QHBoxLayout()
        self.tools_create_box = QGroupBox("Create", self)
        self.tools_create_layout = QVBoxLayout()
        self.tools_scaling_box = QGroupBox("Scaling", self)
        self.tools_scaling_layout = QVBoxLayout()

        tools_edit = dict()
        tools_edit["move"] = QPushButton("Move", self)
        tools_edit["select"] = QPushButton("Select", self)
        tools_edit["edit"] = QPushButton("Edit", self)
        tools_create = dict()
        tools_create["line"] = QPushButton("Line", self)
        tools_create["angle"] = QPushButton("Angle", self)
        tools_create["multi-line"] = QPushButton("Multi-Line", self)
        tools_create["rectangle"] = QPushButton("Rectangle", self)
        tools_create["ellipse"] = QPushButton("Ellipse", self)
        tools_create["polygon"] = QPushButton("Polygon", self)
        tools_scaling = dict()
        tools_scaling["scale"] = QPushButton("Identify Scaling", self)
        tools_scaling["scale"].clicked.connect(self.automatic_scaling)

        # merge dicts
        self.tools = tools_edit | tools_create | tools_scaling | image_tools

        # create button group
        self.tools_buttons = QButtonGroup(self)
        for i, tool in enumerate(self.tools.values()):
            tool.setCheckable(True)
            self.tools_buttons.addButton(tool, id=i)
        self.tools_buttons.button(0).setChecked(True)

        # create edit layout
        for tool in tools_edit.values():
            self.tools_edit_layout.addWidget(tool)
        self.tools_edit_box.setLayout(self.tools_edit_layout)
        self.layout.addWidget(self.tools_edit_box)

        tools_create_list = tools_create.values()
        # create 3 columns for the create tools
        for i, t in enumerate(tools_create_list):
            if i % 3 == 0:
                tools_create_layout = QHBoxLayout()
                self.tools_create_layout.addLayout(tools_create_layout)
            tools_create_layout.addWidget(t)
        self.tools_create_box.setLayout(self.tools_create_layout)
        self.layout.addWidget(self.tools_create_box)

        # select microscope
        select_micop_layout = QHBoxLayout()
        self.micop_label = QLabel("Select Microscope:", self)
        select_micop_layout.addWidget(self.micop_label)

        self.dd_select_micop = QComboBox(self)
        self.dd_select_micop.addItems(self.data_handler.micros_db.keys())
        self.dd_select_micop.setCurrentText(
            self.main_window.settings.value("ui/microscope")
        )
        self.dd_select_micop.currentTextChanged.connect(self.microscope_changed)
        select_micop_layout.addWidget(self.dd_select_micop)
        self.tools_scaling_layout.addLayout(select_micop_layout)

        scaling_vertical_layout = QHBoxLayout()
        self.scaling_direction_dd = QComboBox(self)
        self.scaling_direction_dd.addItems(["horizontal", "vertical"])
        scaling_vertical_layout.addWidget(self.scaling_direction_dd)
        
        threshold_label = "Threshold: "
        self.scaling_label = QLabel(threshold_label, self)
        self.scaling_label.setFixedWidth(QFontMetrics(self.scaling_label.font()).horizontalAdvance(threshold_label))
        scaling_vertical_layout.addWidget(self.scaling_label)
        
        self.scaling_threshold_edit = QLineEdit(self, placeholderText=str(self.DEFAULT_THRESHOLD))
        self.scaling_threshold_edit.setValidator(QIntValidator(0, 255, self))
        
        # Calculate the width required to display "255"
        font_metrics = QFontMetrics(self.scaling_threshold_edit.font())
        width = font_metrics.horizontalAdvance("255")*2  # Adding some padding
        self.scaling_threshold_edit.setFixedWidth(width)
        scaling_vertical_layout.addWidget(self.scaling_threshold_edit)
        
        scaling_vertical_layout.addWidget(tools_scaling["scale"])
        self.tools_scaling_layout.addLayout(scaling_vertical_layout)

        self.tools_scaling_box.setLayout(self.tools_scaling_layout)
        self.layout.addWidget(self.tools_scaling_box)
        self.tools_scaling_box.setLayout(self.tools_scaling_layout)

        # scaling

        self.scaling_factor = None

        posDouble = QDoubleValidator()
        posDouble.setBottom(0)

        self.pixel_edit = QLineEdit(self, placeholderText="Enter value")
        self.pixel_edit.setValidator(posDouble)
        self.pixel_edit.textChanged.connect(self.update_scaling)
        self.length_edit = QLineEdit(self, placeholderText="Enter value")
        self.length_edit.setValidator(posDouble)
        self.length_edit.textChanged.connect(self.update_scaling)

        scaling = QHBoxLayout()
        scaling.addWidget(self.pixel_edit)
        scaling.addWidget(QLabel("px    :  "))
        scaling.addWidget(self.length_edit)

        self.units_dd = QComboBox()
        self.units_dd.setEditable(True)
        self.units_dd.addItems(self.data_handler.units.keys())
        # setting for default unit
        self.units_dd.setMinimumWidth(60)
        self.units_dd.setCurrentIndex(3)
        self.units_dd.currentTextChanged.connect(self.units_changed)
        scaling.addWidget(self.units_dd)

        self.tools_scaling_layout.addLayout(scaling)
        self.tools_scaling_box.setLayout(self.tools_scaling_layout)
        self.layout.addWidget(self.tools_scaling_box)

        # visibility settings
        # self.visibility_box = QGroupBox("Visibility", self)
        # self.visibility_layout = QHBoxLayout()
        # self.visibility_box.setLayout(self.visibility_layout)
        # self.layout.addWidget(self.visibility_box)

        # self.visibility_button = QPushButton("Show all", self)
        # self.visibility_button.clicked.connect(self.vispy_canvas.show_all_objects_w_undo)
        # self.visibility_layout.addWidget(self.visibility_button)

        # self.visibility_button = QPushButton("Hide all", self)
        # self.visibility_button.clicked.connect(self.vispy_canvas.hide_all_objects_w_undo)
        # self.visibility_layout.addWidget(self.visibility_button)

        # selected object
        self.selected_object_box = QGroupBox("Selected Object", self)
        self.selected_object_layout = QVBoxLayout()

        self.selected_structure_layout = QHBoxLayout()

        self.structure_dd = QComboBox(self)
        self.structure_dd.setEditable(True)
        self.structure_dd.setPlaceholderText("Enter structure name")
        self.structure_dd.currentTextChanged.connect(self.structure_dd_changed)
        self.selected_structure_layout.addWidget(self.structure_dd)

        # rename button
        self.rename_button = QPushButton("Rename Structure")
        self.rename_button.clicked.connect(self.rename_structure_ask)
        self.selected_structure_layout.addWidget(self.rename_button)
        self.selected_object_layout.addLayout(self.selected_structure_layout)

        # object list
        self.object_list = QListWidget()
        self.object_list.itemClicked.connect(self.list_object_selected)
        self.selected_object_layout.addWidget(self.object_list)

        self.selected_object_table = QTableWidget()
        # run, when cell content changes via GUI
        self.selected_object_table.itemChanged.connect(self.cell_content_changed)
        self.selected_object_table.setRowCount(0)
        self.set_selected_object_table_columns()

        #  disable changing
        # self.selected_object_table.setEditTriggers(QTableWidget.EditTrigger.NoEditTriggers)

        # do not hide for column headers
        self.selected_object_table.horizontalHeader().hide()
        self.selected_object_table.verticalHeader().hide()
        self.selected_object_table.resizeColumnsToContents()
        # self.selected_object_table.horizontalHeader().setStretchLastSection(True)
        self.selected_object_table.horizontalHeader().setSectionResizeMode(0, QHeaderView.ResizeMode.Stretch)
        

        self.selected_object_layout.addWidget(self.selected_object_table)
        self.selected_object_box.setLayout(self.selected_object_layout)
        self.layout.addWidget(self.selected_object_box)

        # add empty space
        self.layout.addStretch()

        # Overview
        self.openSaveWindow = QPushButton("Show Measurements", self)
        self.layout.addWidget(self.openSaveWindow)
        self.openSaveWindow.clicked.connect(self.parent.open_data_page)

        # Save
        self.openSaveWindow = QPushButton("Save", self)
        self.layout.addWidget(self.openSaveWindow)
        self.openSaveWindow.clicked.connect(self.open_save_window)

        self.setLayout(self.layout)
        
    def set_selected_object_table_columns(self):
        if self.main_window.settings.value("ui/show_both_scaling", type=bool):
            self.selected_object_table.setColumnCount(5)
        else:
            self.selected_object_table.setColumnCount(3)

    def update_object_table(self, object=None):
        """This methods updates the object table with the properties of the selected object.

        Args:
            object: the selected object
        """
        if object is None:
            object = self.vispy_canvas.get_selected_object()
            
        # selection table
        self.clear_object_table()
        if object is None:
            return
        
        self.update_full_table = True # to prevent triggering cell_content_changed

        props = object.output_properties()
        self.selected_object_table.setRowCount(len(props.keys()))
        for i, key in enumerate(props):
            self.selected_object_table.setItem(i, 0, QTableWidgetItem(key))

            value, unit = props[key]
            # if there is a conversion possible
            if self.scaling_factor is not None:
                scaled_length = 1 * np.array(value)
                if key in ["length", "area", "radius", "width", "height", "center"]:
                    scaled_length *= self.scaling_factor
                    exponent = unit[-1] if unit[-1] in ["²", "³"] else ""
                    self.selected_object_table.setItem(
                        i, 2, QTableWidgetItem(self.units_dd.currentText() + exponent)
                    )
                else:
                    self.selected_object_table.setItem(i, 2, QTableWidgetItem(unit))
                self.selected_object_table.setItem(
                    i, 1, QTableWidgetItem(str(scaled_length))
                )
            else:  # to attach flags so you cant edit these columns
                self.selected_object_table.setItem(i, 1, QTableWidgetItem(str(value)))
                self.selected_object_table.setItem(i, 2, QTableWidgetItem(unit))

            # if setting selected that pixels should be shown too
            if self.main_window.settings.value("ui/show_both_scaling", type=bool) and self.scaling_factor is not None:
                self.selected_object_table.setItem(i, 3, QTableWidgetItem(str(value)))
                self.selected_object_table.setItem(i, 4, QTableWidgetItem(unit))
            else:
                self.selected_object_table.setItem(i, 3, QTableWidgetItem(""))
                self.selected_object_table.setItem(i, 4, QTableWidgetItem(""))

        self.selected_object_table.resizeColumnsToContents()
        # make only column 3 editable
        for i in range(self.selected_object_table.rowCount()):
            for j in range(self.selected_object_table.columnCount()):
                if j != 3 and j != 1:
                    item = self.selected_object_table.item(i, j)
                    if item:
                        item.setFlags(item.flags() & Qt.ItemFlag.ItemIsEnabled)

        self.update_full_table = False

    # Slot function that gets called when an item's content changes
    def cell_content_changed(self, item: QTableWidgetItem):
        # in case we update everything, we do not want to trigger this function
        if self.update_full_table:
            return

        # item is the QTableWidgetItem that was changed
        row, column = item.row(), item.column()
        new_value = item.text()

        sel_object = self.vispy_canvas.get_selected_object()
        obj_property = list(sel_object.output_properties().keys())[row]
        old_value = sel_object.output_properties()[obj_property][0]

        if isinstance(old_value, np.ndarray):
            old_type = old_value.dtype
            # load numpy array from string
            new_value = np.array(
                [float(val) for val in new_value.strip("[]").split(" ") if val],
                dtype=old_type,
            )
        else:
            old_type = type(old_value)
            new_value = old_type(new_value)

        # only allow changes in the data value columns
        # this is only the case if the setting for showing both scaling is enabled
        if column == 3:
            self.vispy_canvas.update_object_property_w_undo(
                sel_object, obj_property, new_value, 
                old_value=old_value
            )
        
        if column == 1:
            self.vispy_canvas.update_object_property_w_undo(
                sel_object, obj_property, new_value, 
                self.scaling_factor if self.scaling_factor is not None else 1, 
                old_value=old_value
            )
            

    def select_file(self, file_path=None):
        if file_path is None:
            file_path = self.openFileNameDialog()
        if file_path:
            file_path = Path(file_path)

            self.data_handler.open_file(file_path)

    def openFileNameDialog(self):
        file_name, _ = QFileDialog.getOpenFileName(
            self, "Select Image", "", "All Files (*);;Python Files (*.py)"
        )
        if file_name:
            return file_name

    def open_save_window(self):
        if not self.vispy_canvas.start_state:
            self.save_window = SaveWindow(parent=self)
            self.save_window.show()

    def get_microscope(self) -> Microscope:
        return self.data_handler.micros_db[self.dd_select_micop.currentText()]()

    def automatic_scaling(self):
        # only when image is loaded
        if not self.vispy_canvas.start_state:
            # get seedPoint from micros_db
            try:
                seed_points = self.get_microscope().seed_points
                orientation = self.get_microscope().orientation
                if orientation is not None:
                    self.scaling_direction_dd.setCurrentText(orientation)
                # only actually try to find scaling bar, when data is given by database
                if seed_points is not None:
                    self.vispy_canvas.find_scale_bar_width_w_undo(
                        seed_point_percentage=seed_points, threshold=self.get_threshold(),
                        direction=self.scaling_direction_dd.currentText()
                    )
            except Exception as e:
                self.main_window.raise_error(
                    "Something went wrong while trying to identify scaling bar: "
                    + str(e)
                )

    def update_scaling(self):
        length = self.length_edit.text()
        pixels = self.pixel_edit.text()
        if length != "" and pixels != "":
            self.scaling_factor = float(length.replace(",", "")) / float(
                pixels.replace(",", "")
            )
        else:
            self.scaling_factor = None
        self.units_changed()

    def reset_scaling(self):
        self.units_dd.setCurrentIndex(3)
        self.pixel_edit.setText("")
        self.length_edit.setText("")
        self.units_changed()
        
    def microscope_changed(self):
        threshold = self.get_microscope().threshold
        self.set_threshold(threshold)
        
    def set_threshold(self, threshold):
        if threshold is None:
            self.scaling_threshold_edit.setText("")
        else:
            self.scaling_threshold_edit.setText(str(threshold))
        
    def get_threshold(self):
        threshold = self.scaling_threshold_edit.text()
        return int(threshold) if threshold else self.DEFAULT_THRESHOLD

    def units_changed(self):
        self.selected_object_table.setHorizontalHeaderItem(
            1, QTableWidgetItem(self.units_dd.currentText())
        )
        self.vispy_canvas.selection_update()
        self.right_ui.update_intensity_plot()

    def update_structure_dd(self):
        self.structure_dd.clear()
        if self.data_handler.drawing_data.keys():
            self.structure_dd.addItems(sorted(self.data_handler.drawing_data.keys()))
            self.structure_dd.setCurrentIndex(0)

        self.structure_dd_changed()

    def add_to_structure_dd(self, name):
        if name not in [
            self.structure_dd.itemText(i) for i in range(self.structure_dd.count())
        ]:
            self.structure_dd.addItem(name)

        self.structure_dd.setCurrentText(name)

    def remove_from_structure_dd(self, name):
        names = [
            self.structure_dd.itemText(i) for i in range(self.structure_dd.count())
        ]
        if name in names:
            self.structure_dd.removeItem(names.index(name))
            self.structure_dd.setCurrentIndex(0)

    def structure_dd_changed(self):
        self.update_object_list()

    def update_object_list(self):
        self.object_list.clear()
        if self.structure_dd.currentText() != "":
            if self.structure_dd.currentText() in self.data_handler.drawing_data:
                objects = self.data_handler.drawing_data[
                    self.structure_dd.currentText()
                ]
                types = [type(obj).__name__ for obj in objects]
                self.object_list.insertItems(0, types)

    def clear_object_table(self):
        self.selected_object_table.clearContents()
        self.selected_object_table.setRowCount(0)

    def list_object_selected(self):

        structure = self.structure_dd.currentText()
        index = self.object_list.currentRow()

        object = self.data_handler.drawing_data[structure][index]
        self.vispy_canvas.select(object)

    def rename_structure_ask(self):
        structure = self.structure_dd.currentText()

        if structure:

            new_structure, ok = QInputDialog.getText(
                self,
                "Rename",
                f"Do you want to rename the structure: {structure}?\n",
                text=structure,
            )
            if ok:
                # After the user closes the QInputDialog, you can get the text from the QLineEdit
                self.rename_structure_w_undo(structure, new_structure)
                    
    def rename_structure(self, structure, new_structure):
        status = self.data_handler.rename_structure(structure, new_structure)
        if status:
            self.update_structure_dd()
            if self.structure_dd.currentText() == structure:
                self.structure_dd.setCurrentText(new_structure)
            self.structure_dd_changed()
            
    def rename_structure_w_undo(self, structure_name, new_structure_name):
        command = RenameStructureCommand(self,
                                         structure_name, new_structure_name)
        self.main_window.undo_stack.push(command)       

class RenameStructureCommand(QUndoCommand):
    def __init__(self, main_ui, structure_name, new_structure_name):
        super().__init__()
        self.main_ui = main_ui
        self.structure_name = new_structure_name
        self.old_structure_name = structure_name
        
    def undo(self):
        self.main_ui.data_handler.logger.debug(f"Undo rename structure: {self.structure_name} -> {self.old_structure_name}")
        self.main_ui.rename_structure(self.structure_name, self.old_structure_name)

    def redo(self):
        self.main_ui.data_handler.logger.debug(f"Redo rename structure: {self.old_structure_name} -> {self.structure_name}")
        self.main_ui.rename_structure(self.old_structure_name, self.structure_name)



# absolute imports
from pathlib import Path
from PySide6.QtWidgets import QFileDialog, QMessageBox
from PySide6.QtGui import QGuiApplication, QUndoCommand, QImage
import pickle
import os
from subprocess import Popen
from sys import platform
from copy import deepcopy
import logging
import numpy as np
import cv2

# you need this import as they are implicitly used
from .drawable_objects import EditRectVisual, EditLineVisual, EditEllipseVisual
from .data.microscopes import load_microscopes
from .windows import ImageWindow


class DataHandler:
    img_data = None
    img_byte_stream = None
    file_path: Path | None = None

    main_window: dict | None = None
    drawing_data: dict

    units: dict
    logger: logging.Logger | None = None

    def __init__(self, logger=None):

        self.micros_db = load_microscopes()

        if logger is not None:
            self.logger = logger
        else:
            logging.basicConfig(
                level=logging.CRITICAL,
                format="%(asctime)s  %(levelname)-10s %(name)s: %(message)s",
            )
            self.logger = logging.getLogger("Measury")
            self.logger.setLevel(logging.CRITICAL)

        # drawing data is a dictionary of the form:
        # {
        #   'object 1': [measured_line, measured_line, ...],
        #   'object 2': [measured_circle, measured_circle, ...],
        #   ...
        # }
        self.drawing_data = dict()
        self.units = dict(
            fm=1e-15,
            pm=1e-12,
            Å=1e-10,
            nm=1e-9,
            µm=1e-6,
            mm=1e-3,
            cm=1e-2,
            m=1,
            km=1e3,
        )

    def save_object(self, structure_name, object):
        """save object into storage structure

        Args:
            structure_name (string): _description_
            object_data (_type_): _description_
        """
        if structure_name == "" or structure_name.isspace():
            structure_name = self.generate_output_name()

        # if not in output dict then add an empty list
        if structure_name not in self.drawing_data.keys():
            self.drawing_data[structure_name] = list()
        self.drawing_data[structure_name].append(object)
        self.logger.info(f"object {object} saved in {structure_name} in drawing_data")

        return structure_name

    def delete_object(self, object):
        """delete object from the storage dict

        Args:
            object_data (class:object): the object that should be deleted
        """
        for object_name, object_list in list(self.drawing_data.items()):
            if object in object_list:
                object_list.remove(object)
            # in case it is the controlpoints
            if object.parent in object_list:
                object_list.remove(object.parent)
                object = object.parent
            # If the list is empty after removal, delete the key from the dictionary
            if not object_list:
                self.main_window.main_ui.remove_from_structure_dd(object_name)
                del self.drawing_data[object_name]
                if self.main_window.main_ui.structure_dd.currentText() == object_name:
                    self.main_window.main_ui.structure_dd.setCurrentText("")

    def delete_all_objects(self):
        self.main_window.vispy_canvas.unselect()
        for structure_list in list(self.drawing_data.values()):
            for object in structure_list:
                object.delete()

        self.drawing_data = dict()
        self.main_window.main_ui.update_structure_dd()
        self.logger.info("Deleted all measurements")

    def delete_all_objects_w_undo(self):
        command = DeleteAllObjectsCommand(self)
        self.main_window.undo_stack.push(command)

    def find_object(self, object):
        if self.drawing_data:
            for k, val in self.drawing_data.items():
                if object in val:
                    return k, val.index(object)
            raise LookupError("Object could not be found in drawing_data")

    def generate_output_name(self):
        # return string that is not in the drawing_data keys and increases each time
        i = 1
        while f"structure_{i:03d}" in self.drawing_data.keys():
            i += 1
        return f"structure_{i:03d}"

    def open_file_location(self, path: Path):
        try:
            if path is not None:
                # check if file exists
                if not path.exists():
                    raise FileNotFoundError(f"File not found: {path}")
                
                if platform == "win32":
                    # For Windows
                    Popen(["explorer", "/select,", str(path)])
                elif platform == "darwin":
                    # For MacOS
                    Popen(["open", "-R", str(path)])
                else:
                    # For Linux
                    Popen(["xdg-open", str(path)])
        except Exception as error:
            self.main_window.raise_error(f"Could not open file location: {path.parent}:\n\n{error}")

    def save_file_dialog(
        self, file_name="measury.msry", extensions="Measury Files (*.measury *.msry)"
    ):
        filename, _ = QFileDialog.getSaveFileName(
            self.main_window.main_ui, "Save", file_name, extensions
        )
        if filename:
            return filename

    def save_storage_file(self, save_image=True, **kwargs):
        

        filename = self.save_file_dialog(
            file_name=str(self.file_path.with_suffix(".msry"))
        )
        if filename is None:
            return
        
        self.file_path = Path(filename)
        
        self.logger.info(f"saving storage file: {filename}")

        structure_data = dict()
        for key, val in self.drawing_data.items():
            structure_data[key] = [[obj.__class__, deepcopy(obj.save())] for obj in val]

        if save_image:
            scaling = (
                self.main_window.main_ui.pixel_edit.text(),
                self.main_window.main_ui.length_edit.text(),
                self.main_window.main_ui.units_dd.currentText(),
                self.main_window.vispy_canvas.scale_bar_params,
            )
            output = (self.img_byte_stream, structure_data, scaling, self.main_window.vispy_canvas.origin)
        else:
            output = (None, structure_data, None)

        # save to file
        with open(filename, "wb") as save_file:
            pickle.dump(output, save_file, pickle.HIGHEST_PROTOCOL, **kwargs)

        self.main_window.main_ui.save_window.close()
        self.main_window.vispy_canvas.update_image()

        
    def load_into_view(self, drawing_data, vispy_instance):
        """
        Loads the drawing data into the view and the drawing_data dictionary.

        Args:
            drawing_data (dict): A dictionary containing the drawing data.
            vispy_instance: An instance of the Vispy class.

        Returns:
            None
        """
        self.delete_all_objects()
        self.drawing_data = drawing_data
        self.logger.info("loading drawing data into view and into drawing_data")
        if drawing_data:
            for val in drawing_data.values():
                for object in val:
                    object.parent = vispy_instance.view.scene

    def load_from_pickle(self, file_path):
        self.logger.info(f"opened storage file: {file_path}")
        with open(file_path, "rb") as file:
            unpickler = CustomUnpickler(file)
            obj = unpickler.load()
        return obj

    def load_storage_file(self, file_path:Path):
        """load .msry data from a file and update the view"""

        loaded_data = self.load_from_pickle(file_path)
        # only 3 elements are default, as it becomes 4 when loaded from file 
        scaling = (None, None, None) # pixel, length, unit
        origin = np.zeros(2)
        if len(loaded_data) == 2:
            img_byte_stream, structure_data = loaded_data
        elif len(loaded_data) == 3:
            img_byte_stream, structure_data, scaling = loaded_data
        elif len(loaded_data) == 4:
            img_byte_stream, structure_data, scaling, origin = loaded_data

        
        question = None

        if self.img_data is None and img_byte_stream is None:
            self.main_window.raise_error(
                "You can't load measurements without loading an image first."
            )
            return

        # there is an image with measurements
        if self.drawing_data and self.img_data is not None:
            if img_byte_stream is None:
                if structure_data:
                    # only measurments / no image
                    question = "Do you want to load new measurements?\n\nThis will remove the current measurements."
                else:
                    # if the file is empty
                    return
            else:
                # image included
                if structure_data:
                    question = "Do you want to load a new image with other measurements?\n\nThis will replace the current image and measurements."
                else:
                    question = "Do you want to load a new image?\n\nThis will replace the current image and remove the measurements."

        reply = QMessageBox.StandardButton.Yes
        if question is not None:
            reply = QMessageBox.warning(
                self.main_window.main_ui,
                "Warning",
                question,
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.No,
            )

        if reply == QMessageBox.StandardButton.Yes:

            self.delete_all_objects()
            self.file_path = file_path
            if img_byte_stream is not None:
                if isinstance(img_byte_stream, bytes):
                    self.img_byte_stream = img_byte_stream
                else:  # for old file format where just the pixel matrix is stored
                    # Step 1: Encode the image data to a specific format
                    self.logger.warning("old file format detected")
                    success, encoded_image = cv2.imencode(
                        ".png", img_byte_stream, [int(cv2.IMWRITE_PNG_COMPRESSION), 5]
                    )

                    if success:
                        # Step 2: Convert the encoded image to a byte stream
                        self.img_byte_stream = encoded_image.tobytes()

                self.logger.debug(
                    "set data_handler.img_byte_stream to loaded file data"
                )

            if structure_data:
                for key, val in structure_data.items():
                    for obj_type, obj_data in val:
                        new_object = obj_type(
                            settings=self.main_window.settings,
                            **obj_data,
                            parent=self.main_window.vispy_canvas.view.scene,
                        )
                        self.main_window.vispy_canvas.create_new_object(new_object, structure_name=key)

            if scaling[0] is not None:
                self.main_window.main_ui.pixel_edit.setText(str(scaling[0]))
                self.main_window.main_ui.length_edit.setText(str(scaling[1]))
                # find the index of the unit in the dropdown and set it
                index = self.main_window.main_ui.units_dd.findText(scaling[2])
                if index == -1:
                    self.main_window.main_ui.units_dd.addItem(scaling[2])
                    self.main_window.main_ui.units_dd.setCurrentIndex(
                        self.main_window.main_ui.units_dd.count() - 1
                    )
                else:
                    self.main_window.main_ui.units_dd.setCurrentIndex(index)
                if len(scaling) > 3:
                    self.main_window.vispy_canvas.find_scale_bar_width(*scaling[3])
                    
                self.main_window.main_ui.units_changed()
            
            self.main_window.vispy_canvas.set_origin(origin, move_objects=False)

            self.main_window.main_ui.update_structure_dd()
            self.main_window.vispy_canvas.update_image()

    def open_file(self, file_path: str | Path | None):

        try:
            if file_path:
                file_path = Path(file_path)

                # check if it is a storage file
                if file_path.suffix in self.main_window.settings.value(
                    "misc/file_extensions"
                ):
                    self.load_storage_file(file_path)

                # just assume it is an image file
                else:
                    reply = QMessageBox.StandardButton.Yes
                    if self.img_data is not None:
                        reply = QMessageBox.warning(
                            self.main_window.main_ui,
                            "Warning",
                            "Do you want to load a new image?\n\nThis will replace the current image and remove the measurements.",
                            QMessageBox.StandardButton.Yes
                            | QMessageBox.StandardButton.No,
                            QMessageBox.StandardButton.No,
                        )

                    if reply == QMessageBox.StandardButton.Yes:
                        self.file_path = file_path
                        self.img_byte_stream = file_path.read_bytes()
                        self.delete_all_objects()
                        self.main_window.main_ui.reset_scaling()
                        self.main_window.vispy_canvas.update_image()

        except Exception as error:
            self.main_window.raise_error(f"Could not open file: {file_path}: {error}")

    def open_image_editor(self, image:QImage):
        image_editor = ImageWindow(image, parent=self.main_window)
        # image_editor.show()
        # wait until the window is closed
        image_editor.exec()
        
        return image_editor.image

    def open_image_from_clipboard(self):

        reply = QMessageBox.StandardButton.Yes
        try:
            if self.img_data is not None:
                reply = QMessageBox.warning(
                    self.main_window.main_ui,
                    "Warning",
                    "Do you want to load a new image?\n\nThis will replace the current image and remove the measurements.",
                    QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                    QMessageBox.StandardButton.No,
                )

            if reply == QMessageBox.StandardButton.Yes:

                clipboard = QGuiApplication.clipboard()

                # if it is a file path, open it like a file
                if clipboard.mimeData().hasUrls():
                    file_path = clipboard.mimeData().urls()[0].toLocalFile()
                    self.open_file(file_path)
                    return

                self.file_path = Path("clipboard")
                self.delete_all_objects()
                self.main_window.main_ui.reset_scaling()

                clipboard_data = clipboard.image()
                width = clipboard_data.width()
                height = clipboard_data.height()

                # open image in another window to allow formating
                clipboard_data = self.open_image_editor(clipboard_data)

                # Check if the image is in the correct format
                if clipboard_data.format() != QImage.Format.Format_RGB32:
                    self.logger.info(f"Converting clipboard image from {clipboard_data.format()} to ARGB32 format")
                    clipboard_data = clipboard_data.convertToFormat(QImage.Format.Format_RGB32)
                
                # Get the pointer to the pixel data
                ptr = clipboard_data.constBits()

                # Convert the pixel data to a byte stream
                img_data = np.array(ptr).reshape((height, width, 4))

                # Step 1: Encode the image data to a specific format (e.g., '.png')
                success, encoded_image = cv2.imencode(".png", img_data
                                                      #, [int(cv2.IMWRITE_PNG_COMPRESSION), 5]
                                                     )

                if success:
                    # Step 2: Convert the encoded image to a byte stream
                    self.img_byte_stream = encoded_image.tobytes()

                    self.main_window.vispy_canvas.update_image()
                else:
                    raise Exception("Could not encode image data")

        except Exception as error:
            self.logger.error(f"Could not open from clipboard:\n{error}")
            self.main_window.raise_error(f"Could not open from clipboard: {error}")

    def calculate_results(self):
        """calculate the average and standard deviation of
        the measurements for each structure in drawing_data
        """

        results = dict()
        for structure_name, object_list in self.drawing_data.items():
            results[structure_name] = dict()
            
            # test if it is a multiline
            if isinstance(object_list[0], EditLineVisual):
                num_points = [len(obj.coords) for obj in object_list]
                print(num_points)
                # check if all are the same
                if not all(n == num_points[0] for n in num_points):
                    results[structure_name]["lines have different number of points: "] = (np.mean(num_points), np.std(num_points), "")
                    continue
                # elif num_points[0] == 0 or num_points[0] > 3:
                #     results[structure_name][""] = (num_points[0], None, "")
                #     continue
                
            props = object_list[0].output_properties()
            for prop, value in props.items():
                data = [obj.output_properties()[prop][0] for obj in object_list]
                unit = value[1]
                if self.main_window.main_ui.scaling_factor is not None:
                    if prop in [
                        "length",
                        "area",
                        "radius",
                        "width",
                        "height",
                        "center",
                    ]:
                        data = [
                            d * self.main_window.main_ui.scaling_factor for d in data
                        ]
                        exponent = unit[-1] if unit[-1] in ["²", "³"] else ""
                        unit = (
                            self.main_window.main_ui.units_dd.currentText() + exponent
                        )

                if len(data) == 1:
                    results[structure_name][prop] = (np.mean(data), None, unit)
                else:
                    results[structure_name][prop] = (
                        np.mean(data),
                        np.std(data) / np.sqrt(len(object_list)),
                        unit,
                    )
        return results

    def calculate_results_string(self):
        results = self.calculate_results()
        # sort dict
        results = dict(sorted(results.items()))
        results_string = ""
        for structure_name, structure_results in results.items():
            results_string += f"{structure_name}:\n"
            for prop, prop_results in structure_results.items():
                results_string += f"\t{prop}:\t "
                # value with or without std
                value_str = f"{prop_results[0]:.1f}"
                if prop_results[1] is not None:
                    value_str = f"({value_str} ± {prop_results[1]:.1f})"
                results_string += value_str
                # adding unit
                results_string += f" {prop_results[2]}\n"
            results_string += "\n"
        return results_string

    def save_measurements_file(self):
        filename = self.save_file_dialog(
            file_name=str(self.file_path.with_suffix(".meas")),
            extensions="Measurement File (*.meas)",
        )
        if filename is None:
            return

        with open(filename, "wb") as save_file:
            # write to text file
            save_file.write(self.calculate_results_string().encode("utf-8"))

    def rename_structure(self, structure, new_structure):
        if new_structure in self.drawing_data:
            self.main_window.raise_error("This structure already exists!")
        elif new_structure == "" or new_structure.isspace():
            self.main_window.raise_error("This name is not valid")
        else:
            self.drawing_data[new_structure] = self.drawing_data[structure]
            del self.drawing_data[structure]
            return True

        return False


class DeleteAllObjectsCommand(QUndoCommand):
    def __init__(self, data_handler: DataHandler):
        super().__init__()
        self.data_handler = data_handler
        self.main_window = self.data_handler.main_window
        self.old_drawing_data = dict(data_handler.drawing_data)  # Save the old state

    def undo(self):
        # Restore the old state
        self.data_handler.logger.info("Undoing delete all objects")
        self.data_handler.drawing_data = self.old_drawing_data
        self.main_window.main_ui.update_structure_dd()
        self.data_handler.load_into_view(
            self.data_handler.drawing_data, self.main_window.vispy_canvas
        )
        self.data_handler.logger.info("Restored all measurements")

    def redo(self):
        # Delete all objects
        self.data_handler.delete_all_objects()


class CustomUnpickler(pickle.Unpickler):
    def find_class(self, module, name):

        # if drawable_objects is in the module name,
        # then replace everything before drawable_objects
        # with measury. This allows for the loading of old files
        if "drawable_objects" in module:
            module = "measury.drawable_objects"
        return super().find_class(module, name)
